@@ -16,7 +16,7 @@
 # Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 # MA 02111-1307 USA, or see http://www.gnu.org/.
 
-VERSION = "1.10"
+VERSION = "1.13"
 import feedparser
 from persister import Persistable, Persister
 import os, time, sha, getopt, sys, re, urlparse, cgi, socket
@@ -40,21 +40,21 @@ def format_time(secs, config):
 	return time.strftime(format, t)
 
 def select_content(contents):
-	"""Select the best content element from an Echo feed."""
-	if type(contents) == str:
-		return contents
-	preferred = ["text/html", "application/xhtml+xml"]
+	"""Given a list of contents with alternative content types, select a
+	preferred version."""
+	types = {"text/html": 30,
+	         "application/xhtml+xml": 20,
+	         "text/plain": 10}
 	cs = []
 	for c in contents:
-		ctype = c["type"]
-		if ctype in preferred:
-			score = preferred.index(ctype)
-			cs.append((score, c["value"]))
+		ctype = c.get("type")
+		if types.has_key(ctype):
+			score = types[ctype]
+		else:
+			score = 0
+		cs.append((score, c["value"]))
 	cs.sort()
-	if len(cs) == 0:
-		return None
-	else:
-		return cs[0][1]
+	return cs[-1][1]
 
 def encode_references(s):
 	"""Encode characters in a Unicode string using HTML references."""
@@ -69,20 +69,44 @@ def encode_references(s):
 	r.close()
 	return v
 
-def sanitise_html(html, baseurl, inline = 0):
+# This list of block-level elements came from the HTML 4.01 specification.
+block_level_re = re.compile(r'^\s*<(p|h1|h2|h3|h4|h5|h6|ul|ol|pre|dl|div|noscript|blockquote|form|hr|table|fieldset|address)[^a-z]', re.I)
+def sanitise_html(html, baseurl, inline, config):
 	"""Attempt to turn arbitrary feed-provided HTML into something
 	suitable for safe inclusion into the rawdog output. The inline
 	parameter says whether to expect a fragment of inline text, or a
 	sequence of block-level elements."""
 	if html is None:
 		return None
+
 	# sgmllib handles "<br/>/" as a SHORTTAG; this workaround from
 	# feedparser.
 	html = re.sub(r'(\S)/>', r'\1 />', html)
 	html = feedparser.resolveRelativeURIs(html, baseurl)
 	p = feedparser.HTMLSanitizer()
 	p.feed(html)
-	return encode_references(p.output())
+	html = p.output()
+
+	if not inline:
+		# If we're after some block-level HTML and the HTML doesn't
+		# start with a block-level element, then insert a <p> tag
+		# before it. This still fails when the HTML contains text, then
+		# a block-level element, then more text, but it's better than
+		# nothing.
+		if block_level_re.match(html) is None:
+			html = "<p>" + html
+
+	if config["tidyhtml"]:
+		import mx.Tidy
+		# mx.Tidy won't accept Unicode strings, so we have to
+		# encode data as UTF-8 for it.
+		output = mx.Tidy.tidy(html.encode("UTF-8"), None, None,
+		                      wrap = 0, numeric_entities = 1)[2]
+		output = output.decode("UTF-8")
+		html = output[output.find("<body>") + 6
+		              : output.rfind("</body>")].strip()
+
+	return encode_references(html)
 
 template_re = re.compile(r'__(.*?)__')
 def fill_template(template, bits):
@@ -213,9 +237,6 @@ class Feed:
 
 		self.etag = p.get("etag")
 		self.modified = p.get("modified")
-		# In the event that the feed hasn't changed, then both channel
-		# and feed will be empty. In this case we return 0 so that
-		# we know not to expire articles that came from this feed.
 
 		self.encoding = p.get("encoding")
 		if self.encoding is None:
@@ -233,8 +254,19 @@ class Feed:
 		if channel.has_key("link"):
 			self.link = self.decode(channel["link"])
 
+		# In the event that the feed hasn't changed, then both channel
+		# and feed will be empty. In this case we return 0 so that
+		# we know not to expire articles that came from this feed.
+		if len(p["items"]) == 0:
+			return 0
+
 		feed = self.url
-		seen_items = 0
+
+		if config["currentonly"]:
+			for (hash, a) in articles.items():
+				if a.feed == feed:
+					del articles[hash]
+
 		sequence = 0
 		for item in p["items"]:
 			title = self.decode(item.get("title"))
@@ -242,11 +274,14 @@ class Feed:
 
 			date = item.get("date_parsed")
 			if date is not None:
-				date = time.mktime(date)
+				try:
+					date = time.mktime(date)
+				except OverflowError:
+					date = None
 
 			description = None
 			if description is None and item.has_key("content"):
-				description = self.decode(select_content(item["content"]))
+				description = self.decode(item["content"])
 			if description is None:
 				description = self.decode(item.get("description"))
 
@@ -258,15 +293,17 @@ class Feed:
 				articles[article.hash].last_seen = now
 			else:
 				articles[article.hash] = article
-			seen_items = 1
 
-		return seen_items
+		return 1
 
 	def decode(self, s):
-		"""Convert a string retrieved from the feed from its original
-		encoding to our target encoding for HTML output."""
+		"""Convert a string or alternatives list retrieved from the
+		feed from its original encoding to our target encoding for HTML
+		output."""
 		if s is None:
 			return None
+		if type(s) != str:
+			s = select_content(s)
 		try:
 			us = s.decode(self.encoding)
 		except ValueError:
@@ -285,10 +322,10 @@ class Feed:
 		else:
 			return self.url
 
-	def get_html_link(self):
-		s = sanitise_html(self.get_html_name(), self.get_baseurl(), 1)
+	def get_html_link(self, config):
+		s = sanitise_html(self.get_html_name(), self.get_baseurl(), 1, config)
 		if self.link is not None:
-			return '<a href="' + self.link + '">' + s + '</a>'
+			return '<a href="' + cgi.escape(self.link) + '">' + s + '</a>'
 		else:
 			return s
 
@@ -356,11 +393,14 @@ class DayWriter:
 
 	def time(self, s):
 		tm = time.localtime(s)
-		if tm[:3] != self.lasttime[:3]:
+		if tm[:3] != self.lasttime[:3] and self.config["daysections"]:
 			self.close(0)
 			self.start_day(tm)
-		if tm[:6] != self.lasttime[:6]:
-			self.close(1)
+		if tm[:6] != self.lasttime[:6] and self.config["timesections"]:
+			if self.config["daysections"]:
+				self.close(1)
+			else:
+				self.close(0)
 			self.start_time(tm)
 		self.lasttime = tm
 
@@ -413,6 +453,11 @@ class Config:
 			"itemtemplate" : "default",
 			"verbose" : 0,
 			"ignoretimeouts" : 0,
+			"daysections" : 1,
+			"timesections" : 1,
+			"tidyhtml" : 0,
+			"sortbyfeeddate" : 0,
+			"currentonly" : 0,
 			}
 
 	def __getitem__(self, key): return self.config[key]
@@ -482,6 +527,16 @@ class Config:
 			self["verbose"] = parse_bool(l[1])
 		elif l[0] == "ignoretimeouts":
 			self["ignoretimeouts"] = parse_bool(l[1])
+		elif l[0] == "daysections":
+			self["daysections"] = parse_bool(l[1])
+		elif l[0] == "timesections":
+			self["timesections"] = parse_bool(l[1])
+		elif l[0] == "tidyhtml":
+			self["tidyhtml"] = parse_bool(l[1])
+		elif l[0] == "sortbyfeeddate":
+			self["sortbyfeeddate"] = parse_bool(l[1])
+		elif l[0] == "currentonly":
+			self["currentonly"] = parse_bool(l[1])
 		elif l[0] == "include":
 			self.load(l[1])
 		else:
@@ -611,7 +666,7 @@ by <a href="mailto:azz@us-lot.org">Adam Sampson</a>.</p>
 <span class="itemfrom">[__feed_title__]</span>
 </p>
 __if_description__<div class="itemdescription">
-<p>__description__</p>
+__description__
 </div>__endif__
 </div>
 
@@ -637,13 +692,19 @@ __if_description__<div class="itemdescription">
 
 		bits["refresh"] = """<meta http-equiv="Refresh" """ + 'content="' + str(refresh) + '"' + """>"""
 
+		article_dates = {}
 		articles = self.articles.values()
+		for a in articles:
+			if config["sortbyfeeddate"]:
+				article_dates[a] = a.get_date() or a.added
+			else:
+				article_dates[a] = a.added
 		numarticles = len(articles)
 		def compare(a, b):
 			"""Compare two articles to decide how they
 			   should be sorted. Sort by added date, then
 			   by feed, then by sequence, then by hash."""
-			i = cmp(b.added, a.added)
+			i = cmp(article_dates[b], article_dates[a])
 			if i != 0:
 				return i
 			i = cmp(a.feed, b.feed)
@@ -668,13 +729,13 @@ __if_description__<div class="itemdescription">
 				break
 
 			count += 1
-			dw.time(article.added)
+			dw.time(article_dates[article])
 
 			itembits = {}
 
 			feed = self.feeds[article.feed]
 			baseurl = feed.get_baseurl()
-			title = sanitise_html(article.title, baseurl, 1)
+			title = sanitise_html(article.title, baseurl, 1, config)
 			if title == "":
 				title = None
 			link = article.link
@@ -683,7 +744,7 @@ __if_description__<div class="itemdescription">
 			if feed.args.has_key("format") and feed.args["format"] == "text":
 				description = "<pre>" + cgi.escape(article.description) + "</pre>"
 			else:
-				description = sanitise_html(article.description, baseurl, 0)
+				description = sanitise_html(article.description, baseurl, 0, config)
 			if description == "":
 				description = None
 
@@ -695,13 +756,17 @@ __if_description__<div class="itemdescription">
 					title = "Link"
 
 			itembits["title_no_link"] = title
+			if link is not None:
+				itembits["url"] = link
+			else:
+				itembits["url"] = ""
 			if link is None:
 				itembits["title"] = title
 			else:
-				itembits["title"] = '<a href="' + link + '">' + title + '</a>'
+				itembits["title"] = '<a href="' + cgi.escape(link) + '">' + title + '</a>'
 
-			itembits["feed_title_no_link"] = sanitise_html(feed.title, baseurl, 1)
-			itembits["feed_title"] = feed.get_html_link()
+			itembits["feed_title_no_link"] = sanitise_html(feed.title, baseurl, 1, config)
+			itembits["feed_title"] = feed.get_html_link(config)
 			itembits["feed_url"] = feed.url
 			itembits["feed_hash"] = short_hash(feed.url)
 			itembits["hash"] = short_hash(article.hash)
@@ -721,6 +786,7 @@ __if_description__<div class="itemdescription">
 
 		dw.close()
 		bits["items"] = f.getvalue()
+		bits["num_items"] = str(numarticles)
 		config.log("Selected ", count, " of ", numarticles, " articles to write")
 
 		f = StringIO()
@@ -732,13 +798,14 @@ __if_description__<div class="itemdescription">
 		feeds.sort(lambda a, b: cmp(a.get_html_name().lower(), b.get_html_name().lower()))
 		for feed in feeds:
 			print >>f, '<tr class="feedsrow">'
-			print >>f, '<td>' + feed.get_html_link() + '</td>'
-			print >>f, '<td><a class="xmlbutton" href="' + feed.url + '">XML</a></td>'
+			print >>f, '<td>' + feed.get_html_link(config) + '</td>'
+			print >>f, '<td><a class="xmlbutton" href="' + cgi.escape(feed.url) + '">XML</a></td>'
 			print >>f, '<td>' + format_time(feed.last_update, config) + '</td>'
 			print >>f, '<td>' + format_time(feed.last_update + feed.period, config) + '</td>'
 			print >>f, '</tr>'
 		print >>f, """</table>"""
 		bits["feeds"] = f.getvalue()
+		bits["num_feeds"] = str(len(feeds))
 
 		s = fill_template(self.get_template(config), bits)
 		if outputfile == "-":
